/**
 ** rmp440TrackTaskCodels.c
 **
 ** Codels called by execution task rmp440TrackTask
 **
 ** Author:
 ** Date:
 **
 **/
#include <math.h>
#include <stdint.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "server/rmp440Header.h"
#include "rmp440Const.h"

#include <rmp440/rmp440.h>

static GENPOS_CART_REF stillRef;

/*------------------------------------------------------------------------
 * TrackSpeed
 *
 * Description:
 *
 * Reports:      OK
 *              S_rmp440_POSTER_NOT_FOUND
 *              S_rmp440_BAD_REF
 *              S_rmp440_CMD_STOP_TRACK
 */

/* findTrackedPoster  -  codel START of TrackSpeed
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
findTrackedPoster(RMP440_TRACK_PARAMS *params, int *report)
{
	char *trackedPoster = (char *)params->posterName.name;
	POSTER_ID posterTrackedId;
	GENPOS_CART_STATE *robot = &SDI_F->robot;

	/* Poster Identifier */
	if (posterFind((char *) trackedPoster, &posterTrackedId) != OK)  {
		*report = S_rmp440_POSTER_NOT_FOUND;
		return(END);
	}
	rmp440DataStrId->trackStr.posterId = (long) posterTrackedId;
	rmp440DataStrId->status.rs_mode = RMP440_TRACK;
	/* Set stillRef to current position of the robot */
	stillRef.x = robot->xRef;
	stillRef.y = robot->yRef;
	stillRef.theta = robot->theta;
	stillRef.v = 0.0;
	stillRef.w = 0.0;
	stillRef.vmax = RMP_DEFAULT_MAXIMUM_VELOCITY;
	stillRef.wmax = RMP_DEFAULT_MAXIMUM_YAW_RATE;
	stillRef.linAccelMax = RMP_DEFAULT_MAXIMUM_ACCEL;
	stillRef.angAccelMax = RMP_DEFAULT_MAXIMUM_YAW_ACCEL;
	switch (params->mode) {
	case GENPOS_TRACK_CONFIG:
		rmp440DataStrId->ref.dataType = GENPOS_POS_AND_SPEED_DATA;
		stillRef.dataType = GENPOS_POS_AND_SPEED_DATA;
		break;
	case GENPOS_TRACK_SPEED:
		rmp440DataStrId->ref.dataType = GENPOS_SPEED_DATA;
		stillRef.dataType = GENPOS_SPEED_DATA;
		break;
	default:		
		*report = S_rmp440_BAD_REF;
		return FAIL;
	}
	stillRef.numRef = -1;

	printf("rmp440 tracking on %s\n", (char *)trackedPoster);
	return(EXEC);
}


static ACTIVITY_EVENT
pumpConfigReference(int *report)
{
	GENPOS_CART_REF rmp440Orders;
	GENPOS_CART_REF *refId = &rmp440DataStrId->ref;
	GENPOS_CART_STATE *configId = &rmp440DataStrId->robot;
	static int prevRef = -2;

	/* Lire les consignes sur le poster externe */
	if (posterRead ((POSTER_ID)rmp440DataStrId->trackStr.posterId, 0,
		(void *) &rmp440Orders,
		sizeof (GENPOS_CART_REF)) == sizeof (GENPOS_CART_REF)) {

		/* Check validity of that  */
		if (rmp440Orders.numRef == -1) {
			if (prevRef != -1) {
				printf("Pilo numRef = -1\n");
			}
			prevRef = -1;
			/* Not a valid reference don't move */
			memcpy(refId, &stillRef, sizeof(GENPOS_CART_REF));
			return (EXEC);
		}
		/* Limit speeds and accelerations */
		/* XXX may be set an error */
		if (rmp440Orders.vmax > RMP_DEFAULT_MAXIMUM_VELOCITY)
			rmp440Orders.vmax = RMP_DEFAULT_MAXIMUM_VELOCITY;
		if (rmp440Orders.wmax > RMP_DEFAULT_MAXIMUM_YAW_RATE)
			rmp440Orders.wmax = RMP_DEFAULT_MAXIMUM_YAW_RATE;
		if (rmp440Orders.linAccelMax > RMP_DEFAULT_MAXIMUM_ACCEL)
			rmp440Orders.linAccelMax = RMP_DEFAULT_MAXIMUM_ACCEL;
		if (rmp440Orders.angAccelMax > RMP_DEFAULT_MAXIMUM_YAW_ACCEL)
			rmp440Orders.angAccelMax = RMP_DEFAULT_MAXIMUM_YAW_ACCEL;

		/* Send reference */
		refId->backFlag = (rmp440Orders.v > 0 ?
				   GENPOS_FORWARD_MOTION :
				   GENPOS_BACKWARD_MOTION);
		refId->v = fabs(rmp440Orders.v);
		refId->vt = 0;
		refId->w = rmp440Orders.w;
		refId->vmax = rmp440Orders.vmax;
		refId->wmax = rmp440Orders.wmax;
		refId->linAccelMax = rmp440Orders.linAccelMax;
		refId->angAccelMax = rmp440Orders.angAccelMax;
		refId->x = rmp440Orders.x;
		refId->y = rmp440Orders.y;
		refId->theta =  angleLimit(rmp440Orders.theta);
		refId->dataType = GENPOS_POS_AND_SPEED_DATA;
		/* store position in case of stop */
		stillRef.x = refId->x;
		stillRef.y = refId->y;
		stillRef.theta = refId->theta;
		prevRef = rmp440Orders.numRef;
		return(EXEC);
	}
	*report = errnoGet();
	return(END);
}

static ACTIVITY_EVENT
pumpSpeedReference(int *report)
{
	GENPOS_CART_SPEED rmp440Orders;
	GENPOS_CART_REF *refId = &rmp440DataStrId->ref;
	GENPOS_CART_STATE *configId = &rmp440DataStrId->robot;

	/* Lire les consignes sur le poster externe */
	if (posterRead ((POSTER_ID)rmp440DataStrId->trackStr.posterId, 0,
		(void *) &rmp440Orders,
		sizeof (GENPOS_CART_SPEED)) == sizeof (GENPOS_CART_SPEED)) {

		/* Verifier la validite des arguments */
		if (rmp440Orders.v > RMP_DEFAULT_MAXIMUM_VELOCITY ||
		    rmp440Orders.w > RMP_DEFAULT_MAXIMUM_YAW_RATE) {

			/* On met une consigne de vitesse nulle sur le point
			 courant. XXXX ouch. */
			refId->v = 0.;
			refId->w = 0.;
			refId->dataType = GENPOS_SPEED_DATA;
			*report = S_rmp440_BAD_REF;
			return (END);
		}

		/* Transmettre la consigne */
		else {
			refId->backFlag = (rmp440Orders.v > 0 ?
			    GENPOS_FORWARD_MOTION : GENPOS_BACKWARD_MOTION);
			refId->v = fabs(rmp440Orders.v);
			/* printf("pumpSpeedRef: %lf\n", refId->v); */
			refId->vt = 0;
			refId->w = rmp440Orders.w;
			refId->vmax = rmp440Orders.vmax;
			refId->wmax = rmp440Orders.wmax;
			refId->linAccelMax = rmp440Orders.linAccelMax;
			refId->angAccelMax = rmp440Orders.angAccelMax;
			refId->dataType = GENPOS_SPEED_DATA;
			return(EXEC);
		}
	} else {
		printf("pumpSpeedReference: posterRead error\n");
	}
	*report = errnoGet();
	return(END);
}

/* pumpReference  -  codel EXEC of TrackSpeed
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
pumpReference(RMP440_TRACK_PARAMS *params, int *report)
{
	GENPOS_TRACK *track = &rmp440DataStrId->trackStr;
	RMP440_STATUS_STR *status = &rmp440DataStrId->status;

	/* Check if mode changed */
	switch (status->rs_mode) {
	case RMP440_EMERGENCY:
		*report = S_rmp440_EMERGENCY_STOP;
		return END;
	case RMP440_MOTORS_OFF:
		*report = S_rmp440_MOTORS_OFF;
		return END;
	}

	switch (track->trackMode) {
	case GENPOS_TRACK_POS:
		return pumpConfigReference(report);
	case GENPOS_TRACK_CONFIG:
		return pumpConfigReference(report);
	case GENPOS_TRACK_SPEED:
		return pumpSpeedReference(report);
	default:
		*report = EXEC_TASK_BILAN(RMP440_MOTIONTASK_NUM);
		return(END);

	}
}

/* smoothStopTrack  -  codel END of TrackSpeed
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
smoothStopTrack(RMP440_TRACK_PARAMS *params, int *report)
{
	GENPOS_CART_REF *refId;	       /* Id structure de consignes */
	GENPOS_CART_STATE *robotId;	/* Id config du robot */
	GENPOS_TRACK *trackId;
	double dtArret;                /* Temps d'arret */
	double dtRetard;               /* Temps de prise en compte de la consigne */
	double dsArret;                /* Longueur d'arret */
	double dtLoco;                 /* Periode d'asserv de la loco */
	double angle;                  /* Moyenne entre angles d'init et d'arret */
	double theta;

	printf("rmp440 smoothStopTrack\n");

	/* Obtenir les pointeurs vers les structures de donnees de la locomotion */
	robotId = &(rmp440DataStrId->robot);
	refId = &rmp440DataStrId->ref;
	trackId = &rmp440DataStrId->trackStr;

	if (trackId->trackMode == GENPOS_TRACK_SPEED) {
		/* Set a null speed and stop the tracking */
		trackId->trackMode = GENPOS_NO_TRACKING;
		refId->dataType = GENPOS_SPEED_DATA;
		refId->x = robotId->xRob;
		refId->y = robotId->yRob;
		refId->theta = robotId->theta;
		refId->v = 0.0;
		refId->w = 0.0;
	} else {
		refId->dataType = GENPOS_POS_AND_SPEED_DATA;

		/* Determination de la periode d'asserv de la loco */
		dtLoco = EXEC_TASK_PERIOD(RMP440_TRACKTASK_NUM);

		/* Temps d'arret minimal (tps de decel) */
		dtArret = MAX (fabs(robotId->v) / RMP_DEFAULT_MAXIMUM_DECEL,
			fabs(robotId->w) / RMP_DEFAULT_MAXIMUM_YAW_ACCEL);

		/* Temps de prise en compte (v = cst) */
		dtRetard =  1 * dtLoco;

		/* Angle final (avec periode de prise en compte) */
		theta = robotId->theta + robotId->w * (dtArret / 2 + dtRetard);

		/* Angle moyen */
		angle = (robotId->theta + theta) / 2;

		/* Longueur d'arret */
		dsArret = robotId->v * (dtArret / 2 + dtRetard);

		/* Prevision position du centre */
		refId->x = robotId->xRob + dsArret * cos (angle);
		refId->y = robotId->yRob + dsArret * sin (angle);

		/* Vitesses de consigne nulles */
		refId->v = 0. ;
		refId->w = 0. ;

		/* Indiquer l'arret du tracking */
	}
	trackId->trackMode = GENPOS_NO_TRACKING;
	rmp440DataStrId->status.rs_mode = RMP440_IDLE;
	posterWrite(RMP440_STATUS_POSTER_ID, 0, &(rmp440DataStrId->status), sizeof(RMP440_STATUS_STR));
	return ETHER;
}
